// vincia09-root: 
// This is test program to check 4-parton phase-space points 
// Matrix elements vs global antenna functions and AP kernels
// With and without polarizations
// Copyright (C) 2019 Peter Skands

// PYTHIA and VINCIA include files
#include "Vincia/Vincia.h"
#include "Vincia/MG4interface.h"
#include "VinciaRoot/VinciaRoot.h"

// Use Pythia8 and Vincia namespaces, for convenience.
using namespace Pythia8; 
using namespace Vincia; 

// Here set whether to impose phase-space restrictions, and what kind
// Parton numbering is 0 = resonance, then final state = 1, 2, 3, 4 
static const bool   restrictPS = false;
static const double r12min = 0.0;
static const double r12max = 1.0;
static const double r23min = 0.0;
static const double r23max = 1.0;
static const double r34min = 0.0;
static const double r34max = 1.0;
static const double r123min = 0.0;
static const double r123max = 1.0;

void listPartons(vector<int> id, vector<Vec4> p, vector<int> h) {
  cout<<setprecision(3);
  for (int i=0; i<(int)p.size(); ++i) {
    cout<<" "<<i<<setw(4)<<id[i];
    cout<<"  h = "<<setw(2)<<h[i];
    cout<<"  (p,E) = "<<setw(9)<<p[i].px()<<setw(9)
        <<p[i].py()<<setw(9)<<p[i].pz()<<setw(9)
        <<p[i].e()<<endl;    
  }
  cout<<setprecision(6);
}

// Aux functions: pT2 and mD2 resolution functions

double qT2(Vec4 p1, Vec4 p2, Vec4 p3) {
  return 2*m2(p1+p2)*m2(p2+p3)/m2(p1+p2+p3); 
}

double mD2(Vec4 p1, Vec4 p2, Vec4 p3) {
  return min(m2(p1+p2),m2(p2+p3));
}

double q2(Vec4 p1, Vec4 p2, Vec4 p3, int evType=1) {
  if (evType == 1) return qT2(p1,p2,p3);
  else return mD2(p1,p2,p3);
}

// Main Program
int main() {

  //************************************************************************
  // Define Pythia 8 generator with plug-in
  Pythia pythia; 
  VinciaPlugin vincia(&pythia);
  pythia.readFile("vincia09.cmnd");

  //************************************************************************

  // Shorthands
  Settings* settingsPtr       = vincia.settingsPtr;
  VinClu* vinCluPtr           = &vincia.vinClu;
  MG4interface* mgInterfacePtr= &vincia.mg4interface;
  AntennaSetFSR* antSetGlbPtr = &vincia.antennaSetFSR;
  Functions* functionsPtr     = &vincia.functions;
  Resolution* resolutionPtr   = &vincia.resolution;

  // Extract settings to be used in the main program.  
  double eCM            = settingsPtr->parm("Beams:eCM");
  double S              = pow(eCM,2);
  int    nEvent         = settingsPtr->mode("Main:numberOfEvents");
  int    nCount         = settingsPtr->mode("Next:numberCount");
  int    verbose        = settingsPtr->mode("Vincia:verbose");
  int    evType         = settingsPtr->mode("Vincia:evolutionType");

  // Use Pythia's spare parameters to define what process we look at

  // Main:spareMode1 = ID of decaying particle
  //     23 : Z0
  //     24 : W-  (normally PDG code for W- is -24, but here take abs)
  //     25 : H0
  //     37 : H-  (normally PDG code for H- is -37, but here take abs)
  int idRes    = abs(settingsPtr->mode("Main:spareMode1"));

  // Main:spareMode2 = Number of quark pairs in final state
  //     0 : pure gluons (only applicable to H decays)
  //     1 : qq + gluons
  //     2 : qqqq 
  int nQQpairs = settingsPtr->mode("Main:spareMode2");

  // Main:spareMode3 could be used for final-state multiplicity. 
  // For now fixed to 4
  int nPartons = 4;

  // Main:spareFlagN : true for left-handed partons
  int h1 = (settingsPtr->flag("Main:spareFlag1")) ? -1 : 1;
  int h2 = (settingsPtr->flag("Main:spareFlag2")) ? -1 : 1;
  int h3 = (settingsPtr->flag("Main:spareFlag3")) ? -1 : 1;
  // The helicity of the last parton is determined by helicity conservation

  //************************************************************************
  // VinciaRoot
  //************************************************************************

  // Create VinciaRoot object and set default behavior to logarithmic y axes
  VinciaRoot vroot(settingsPtr); 
  vroot.init();
  
  // Change default to logarithmic Y axis and switch off ratio panes
  vroot.setLogY();
  vroot.setShowRatioPad("all",false);

  // Tell VinciaRoot how verbose you want it to be
  vroot.setVerbose("all",verbose);
  
  // Set MC names
  string vcVer = num2str(settingsPtr->parm("Vincia:versionNumber"),5);  
  string MCnames;
  MCnames += "Vincia "+vcVer+" + MadGraph 4.426";
  vroot.setMCnames("all",MCnames);

  // Set label for plot legend
  vector<string> legend;
  legend.push_back("Full Phase-Space");
  legend.push_back("y_{23} < 0.01");
  legend.push_back("y_{23} < 0.01 (R)");
  vroot.setLegend("all",legend);

  // Set first update and save event
  string fileNameBase = "v09";
  if (idRes == 23) fileNameBase += "z";
  else if (idRes == 25) fileNameBase += "h";
  if (nQQpairs == 0) fileNameBase+="4g";
  else if (nQQpairs == 1) fileNameBase+="2q2g";
  else if (nQQpairs == 2) fileNameBase+="4q";

  // Set name of output ROOT file
  ostringstream rootFileName;
  rootFileName<<"vincia09.root";
  // Open ROOT file
  TFile rootFile(rootFileName.str().c_str(),"recreate");  
  // Opened root file:
  cout <<"vincia09-root : opened ROOT file "<<rootFileName.str()<<endl;

  //************************************************************************
  // Initialize
  vincia.init();

  //************************************************************************
  // Book Histograms
  //************************************************************************

  // Histogram titles
  string resonance = "Z^{0}";
  if (idRes == 24) resonance = "W^{#pm}";
  else if (idRes == 25) resonance = "H^{0}";
  else if (idRes == 37) resonance = "H^{#pm}";
  string decays = "gggg";
  if (nQQpairs == 1) decays = "qgg#bar{q}";
  else if (nQQpairs == 2) decays = "q#bar{q}q#bar{q}";
  string process = resonance+" #rightarrow "+decays;
  string glbTit  = process+" (Global)";
  string apTit  = process+" (AP)";

  // Histogram bin and range variables
  int nBins;
  double xLo, xHi, yMin, yMax;
  double xLoR = -4.2;
  if (nQQpairs == 1) xLoR = -2.1;

  // Ln(PS/ME) global
  nBins = 21;
  xLo   = xLoR;
  xHi   = abs(xLoR);
  yMin  = 0.001;
  yMax  = 50.;
  vector<TH1D*> VRglb;
  VRglb.push_back(new TH1D("LnRglb",glbTit.c_str(),nBins,xLo,xHi));
  VRglb.push_back(new TH1D("LnRglb1",glbTit.c_str(),nBins,xLo,xHi));
  VRglb.push_back(new TH1D("LnRglb2",glbTit.c_str(),nBins,xLo,xHi));
  double wRglb = nBins/(xHi-xLo);
  VRglb[0]->GetXaxis()->SetTitle("#xi=Log_{10}(PS/ME)");
  VRglb[0]->GetYaxis()->SetTitle("dP/d#xi");
  VRglb[0]->SetMinimum(yMin);
  VRglb[0]->SetMaximum(yMax);
  vroot.createDisplay("Rglb",VRglb[0],360,360);
  for (unsigned int iVar=1; iVar < VRglb.size(); ++iVar) 
    vroot.addTH1DPtr("Rglb",VRglb[iVar]);
  vroot.setLogY("Rglb");
  // add line at zero
  TLine* lineZeroPtr = new TLine(0.,yMin,0.,yMax);
  lineZeroPtr->SetLineColor(17);
  lineZeroPtr->SetLineWidth(2.);
  lineZeroPtr->SetLineStyle(7);
  vroot.addTLine("Rglb",lineZeroPtr);
  vroot.setNormalize("Rglb",true);
  
  // Log10(PS/ME) DGLAP
  xLo   = xLoR;
  xHi   = abs(xLoR);
  vector<TH1D*> VRap;
  VRap.push_back(new TH1D("LnRap",apTit.c_str(),nBins,xLo,xHi));
  VRap.push_back(new TH1D("LnRap1",apTit.c_str(),nBins,xLo,xHi));
  VRap.push_back(new TH1D("LnRap2",apTit.c_str(),nBins,xLo,xHi));
  double wRap = nBins/(xHi-xLo);
  VRap[0]->GetXaxis()->SetTitle("#xi=Log_{10}(PS/ME)");
  VRap[0]->GetYaxis()->SetTitle("dP/d#xi");
  VRap[0]->SetMinimum(yMin);
  VRap[0]->SetMaximum(yMax);
  vroot.createDisplay("Rap",VRap[0],360,360,720,0);
  for (unsigned int iVar=1; iVar < VRap.size(); ++iVar) 
    vroot.addTH1DPtr("Rap",VRap[iVar]);
  vroot.setLogY("Rap");
  vroot.addTLine("Rap",lineZeroPtr);
  vroot.setNormalize("Rap",true);

  // Weight distribution
  xLo   = -8.;
  xHi   = 4.0;
  vector<TH1D*> Vweights;
  Vweights.push_back(new TH1D("Weights","Weights",nBins,xLo,xHi));
  Vweights.push_back(new TH1D("Weights1","Weights",nBins,xLo,xHi));
  Vweights.push_back(new TH1D("Weights2","Weights",nBins,xLo,xHi));
  double wWeights = nBins/(xHi-xLo);
  Vweights[0]->GetXaxis()->SetTitle("w=log_{10}(ME)");
  Vweights[0]->GetYaxis()->SetTitle("dP/dw");
  Vweights[0]->SetMinimum(0.0001);
  Vweights[0]->SetMaximum(10.0);
  vroot.createDisplay("Weights",Vweights[0],200,200,0,360);
  for (unsigned int iVar=1; iVar < Vweights.size(); ++iVar) 
    vroot.addTH1DPtr("Weights",Vweights[iVar]);
  vroot.setLogY("Weights");
  vroot.setNormalize("Weights",true);

  // 2D histograms (still an experimental new feature)
  int n2D = 50;
  double x2lo =  0.;
  double x2hi =  2.5;
  double y2lo =  0.;
  double y2hi =  2.5; 
  TH2D* Hn2D
    = new TH2D("n2D","N",n2D,x2lo,x2hi,n2D,y2lo,y2hi);  
  TH2D* Hr2Dsmooth
    = new TH2D("r2Dsmooth","Smooth Ordering: PS/ME",n2D,x2lo,x2hi,n2D,y2lo,y2hi);  
  TH2D* Hr2Dstrong
    = new TH2D("r2Dstrong","Strong Ordering: PS/ME",n2D,x2lo,x2hi,n2D,y2lo,y2hi);  
  vroot.createDisplay("n2D",Hn2D,400,400);

  //************************************************************************
  // Generate phase-space points
  //************************************************************************

  cout << "Generating "<<nEvent<<" "<<nPartons<<"-parton phase space "
    "point(s)\n";
  int nRejected = 0;
  for (int iEvent=1;iEvent<=nEvent;iEvent++) {

    if (iEvent%(max(1,nEvent/max(nCount,1))) == 0)  {
      cout << " Now begin event " << iEvent;
      if (nRejected > 0) cout<<" [ nRejected = "<<nRejected<<"; f(accept) = "
                             << (1.*iEvent)/(nRejected+iEvent)<<" ]";
      cout<<endl;
    }

    // Set ID of partons, with decaying resonance first
    vector<int> idPartons;
    idPartons.resize(0);
    idPartons.push_back(idRes);
    if (nQQpairs == 0) idPartons.push_back(21);
    else idPartons.push_back(5);
    if (nQQpairs <= 1) {
      idPartons.push_back(21);
      idPartons.push_back(21);
    } else {
      idPartons.push_back(-3);      
      idPartons.push_back(3);
    }
    if (nQQpairs == 0) idPartons.push_back(21);
    else if (idRes == 24 || idRes == 37) idPartons.push_back(-4);
    else idPartons.push_back(-5);

    // Set momenta of partons, with decaying resonance first
    vector<Vec4> pPartons;
    bool accept = false;
    double r12, r23, r34, r123;
    while ( ! accept ) {
      // Convention: p[0][i] is the energy of the i'th parton
      double p[4][8];
      vincia.rambo.genPoint(S,nPartons,p);
      pPartons.resize(0);
      pPartons.push_back(Vec4(0.,0.,0.,eCM));
      for (int iParton=0;iParton<=3;iParton++) { 
        Vec4 VecP;
        VecP.px(p[1][iParton]);
        VecP.py(p[2][iParton]);
        VecP.pz(p[3][iParton]);
        VecP.e(p[0][iParton]);
        pPartons.push_back(VecP);
      }    
      // Phase space restrictions?
      accept = true;
      r12 = m2(pPartons[1]+pPartons[2])/S;
      r23 = m2(pPartons[2]+pPartons[3])/S;
      r34 = m2(pPartons[3]+pPartons[4])/S;
      r123 = m2(pPartons[1]+pPartons[2]+pPartons[3])/S;
      if (restrictPS) {
	if (r12 > r12max || r12 < r12min) accept = false;
	else if (r23 > r23max || r23 < r23min) accept = false;
	else if (r34 > r34max || r34 < r34min) accept = false;
	else if (r123 > r123max || r123 < r123min) accept = false;
      }
      if (! accept) nRejected++;
    }
    
    // Set helicities of partons, with decaying resonance unpolarized
    vector<int> hPartons;
    hPartons.push_back(9);
    hPartons.push_back(h1);
    hPartons.push_back(h2);
    hPartons.push_back(h3);
    // Z and W decays: last one must be flipped. Scalars unflipped.
    if (idRes == 23 || idRes == 24) hPartons.push_back(-h1);
    else hPartons.push_back(h1);

    // Print out phase-space configuration
    if (verbose >= 3) listPartons(idPartons,pPartons,hPartons);

    // Compute MadGraph matrix elements (unpolarized and helicity-dependent)
    double MGME    = mgInterfacePtr->ME2(1,idPartons,pPartons);
    double MGMEpol = mgInterfacePtr->ME2(1,idPartons,pPartons,hPartons);
    Vweights[0]->Fill(log10(MGME),wWeights);
    

    // Verbose output: inform user what the matrix elements are
    if (verbose >= 2) {
      cout<<"\nMadgraph matrix elements for 4-parton state:\n";
      cout<<"    ME("<<setw(3)<<idRes<<" -> ";
      for (int iParton=1; iParton<int(idPartons.size()); ++iParton) {
        cout<<setw(2)<<idPartons[iParton];
        if (iParton < (int)idPartons.size() - 1) cout<<",";
      }
      cout<<") = "<<MGME<<" (unpolarized)"<<endl;
      cout<<"                           = "<<MGMEpol
          <<" (for helicity test configuration ";
      for (int iParton=1; iParton<int(idPartons.size()); ++iParton) {
        cout<<setw(2)<<hPartons[iParton];
        if (iParton < (int)idPartons.size() - 1) cout<<",";
      }
      cout<<")\n\n";
    }
  
    if (verbose >=2) cout<<"Now building shower approximations:\n\n";

    // Determine which sector to use for sector antennae
    unsigned int irSct = 0;
    double QSmin2 = m2(pPartons[0]);
    for (unsigned int ir = 2; ir < pPartons.size(); ir++) {
      // Check if cyclic (last parton = gluon)
      int iK = ir+1;
      if (ir == pPartons.size() -1) {
        if (idPartons[ir] != 21) continue;
        iK = 1;
      }
      int sectorMeasureType = 10;
      if (idPartons[ir] < 0) sectorMeasureType = 8;  
      else if (idPartons[ir] != 21) sectorMeasureType = 9;  
      double QS2 = resolutionPtr->Q2E(pPartons[ir-1],pPartons[ir],pPartons[iK],
			     sectorMeasureType);
      if (verbose >= 2) 
        cout<<"    Sector "<<ir<<" QS = "<<sqrt(QS2)<<endl;
      if (QS2 < QSmin2) {
	QSmin2 = QS2;
	irSct  = ir;
      }
    }

    // Determine polarized and unpolarized antenna and DGLAP approximations
    int antIndx;
    bool swapAB;
    double antGlbSum = 0.;
    double antGlbStrongSum = 0.;
    double PzAPsum   = 0.;
    double antGlbHelSum  = 0.;
    double PzAPhelSum    = 0.;
    for (unsigned int ir = 2; ir < pPartons.size(); ++ir) {

      // Check if last parton is gluon (cyclic) or not (skip)
      int iK = ir+1;
      if (ir == pPartons.size() -1) {
        if (idPartons[ir] != 21) continue;
        iK = 1;
      }

      // Do clustering
      vector<Vec4> pClu;
      vector<int> idClu;      
      vector< vector<int> > hCluVec;
      vinCluPtr->getCluPol(idPartons,pPartons,ir-1,ir,iK,
			   idClu,pClu,antIndx,swapAB,hPartons,hCluVec);      

      // Get unpolarized matrix element for this clustering
      double MEclu = mgInterfacePtr->ME2(1,idClu,pClu);

      // Invariants
      double sijk = m2(pPartons[ir-1]+pPartons[ir]+pPartons[iK]);
      double yij  = m2(pPartons[ir-1]+pPartons[ir])/sijk;
      double yjk  = m2(pPartons[ir]+pPartons[iK])/sijk;
      double y12  = yij;
      double y23  = yjk;
      int    hi   = hPartons[ir-1];
      int    hj   = hPartons[ir];
      int    hk   = hPartons[iK];
      if (swapAB) {
	double yTmp = yij;
	yij = yjk;
	yjk = yTmp;
	int hTmp = hi;
	hi  = hk;
	hk  = hTmp;
      }
      
      // Compute Pimp factor, for global showers
      double q2Now = ( idPartons[ir] == 21 ) ? 
        q2(pPartons[ir-1],pPartons[ir],pPartons[iK],evType) : yjk*sijk;
      double q2Old = sijk;
      for (unsigned int jr = 2; jr < pClu.size(); ++jr) {        
        // Check if last parton is gluon (cyclic) or not (skip)
        int jK = jr+1;
        if (jr == pClu.size() -1) {
          if (idClu[jr] != 21) continue;
          jK = 1;
        }
        if ( idClu[jr] == 21)
          q2Old = q2(pClu[jr-1],pClu[jr],pClu[jK],evType);
        else if (idClu[jr] < 0) 
          q2Old = m2(pClu[jr]+pClu[jK]);
        else 
          q2Old = m2(pClu[jr]+pClu[jr-1]);        
      }
      double Pimp = q2Old/(q2Old+q2Now);      
      double PimpStrong = (q2Old > q2Now) ? 1.0 : 0.0;

      // Compute ARIADNE factor, for g->qq in global showers
      double Pari = 1.0;
      // Deprecated since Vincia v2.2; set =1 here.
      /*
      double sP = sijk;      
      double sN = sijk;
      if (idPartons[ir] < 0) sN = m2(pClu[ir]+pClu[ir+1]);
      else if (idPartons[ir] != 21) sN = m2(pClu[ir-2]+pClu[ir-1]);
      double Pari = 2*sN/(sN+sP);
      */

      // Get unpolarized antenna functions for this clustering
      AntennaFSR* antGlbPtr  = antSetGlbPtr->getAnt(antIndx);
      double antGlb = antGlbPtr->antennaFunction(yij,yjk)/sijk;
      
      // Get unpolarized DGLAP values for this clustering
      // Assign half of gluon splitting to each antenna clustering
      // For quark clusterings, z is energy fraction of quark, 1-z of antiquark
      double PzAPup(0.), PzAPdn(0.);      
      double zUp = pPartons[ir-1].e()/(pPartons[ir].e() + pPartons[ir-1].e());
      double zDn = pPartons[iK].e()/(pPartons[ir].e() + pPartons[iK].e());
      double q2Up = m2(pPartons[ir-1]+pPartons[ir]);
      double q2Dn = m2(pPartons[ir]+pPartons[iK]);
      double ptEvolUp = zUp*(1-zUp)*q2Up;
      double ptEvolDn = zDn*(1-zDn)*q2Dn;
      double PordUp = (q2Old > ptEvolUp) ? 1.0 : 0.0;
      double PordDn = (q2Old > ptEvolDn) ? 1.0 : 0.0;
      if (idPartons[ir] < 0) {
        // Antiquark clustering with downstream quark
        PzAPdn = 0.5*functionsPtr->Pg2qq(zDn)/(yjk*sijk);
      } else if (idPartons[ir] != 21) {
        // Quark clustering with upstream antiquark
        PzAPup = 0.5*functionsPtr->Pg2qq(1.-zUp)/(yjk*sijk);
      } else {
        // Gluon clustering. Upstream side can be either quark or gluon.
        PzAPup = (idPartons[ir-1] != 21) ? functionsPtr->Pq2qg(zUp)/(y12*sijk)
          : 0.5*functionsPtr->Pg2gg(zUp)/(y12*sijk);        
        // Gluon clustering. Downstream side can be either gluon or qbar.
        PzAPdn = (idPartons[iK] != 21) ? functionsPtr->Pq2qg(zDn)/(y23*sijk)
          : 0.5*functionsPtr->Pg2gg(zDn)/(y23*sijk);
      }

      // Compare antenna and AP answers
      if (verbose >= 3) {
	cout<<"\n       Clu "<<ir<<" wME = "<<MEclu<<"  Glb = "<<antGlb<<" ";
	cout<<" AP = "<<PzAPup+PzAPdn
            <<" = "<<setprecision(6)<<PzAPup<<" + "<<PzAPdn
            <<" (zu="<<setprecision(3)<<zUp<<",zd="<<zDn
            <<")"<<setprecision(6)<<endl;
      }

      // Get shower color factor
      double colFac = antGlbPtr->getChargeFac();

      // Sum 
      antGlbSum += colFac * Pimp * Pari * antGlb * MEclu;
      antGlbStrongSum += colFac * PimpStrong * Pari * antGlb * MEclu;
      PzAPsum   += colFac * (PordUp * PzAPup + PordDn * PzAPdn) * MEclu;

      // Now do helicity approximations. Loop over gluon helicity histories
      for (int iHel=0; iHel < (int)hCluVec.size(); ++iHel) {	
	vector<int> hClu = hCluVec[iHel];
	int hA = hClu[ir-1];
	int hB = hClu[ir];
	if (swapAB) {
	  int hTmp = hA;
	  hA = hB;
	  hB = hTmp;
	}
	// Get unpolarized matrix element for this clustering
	double MEcluHel = mgInterfacePtr->ME2(1,idClu,pClu,hClu);
	// Get polarized antenna functions
	double antGlbHel = antGlbPtr->antennaFunction(yij,yjk,0.,0.,0.,
						      hA,hB,hi,hj,hk)/sijk;

	// Get helicity-dependent DGLAP functions
        int hG    = hClu[2];
        int hUp   = hPartons[ir-1];
        int hR    = hPartons[ir];
        int hDn   = hPartons[iK];        
	double PzAPupHel(0.), PzAPdnHel(0.);
        if (idPartons[ir] < 0) {
          // Antiquark clustering with downstream quark
          PzAPdnHel = 0.5*functionsPtr->Pg2qq(zDn,hG,hDn,hR)/(yjk*sijk);
        } else if (idPartons[ir] != 21) {
          // Quark clustering with upstream antiquark
          PzAPupHel = 0.5*functionsPtr->Pg2qq(1.-zUp,hG,hR,hUp)/(yjk*sijk);
        } else {
          // Gluon clustering. Upstream side can be either quark or gluon.
          PzAPupHel = (idPartons[ir-1] != 21) ? 
            functionsPtr->Pq2qg(zUp,hClu[ir-1],hUp,hR)/(y12*sijk)
            : 0.5*functionsPtr->Pg2gg(zUp,hClu[ir-1],hUp,hR)/(y12*sijk);        
          // Gluon clustering. Downstream side can be either gluon or qbar.
          PzAPdnHel = (idPartons[iK] != 21) ? 
            functionsPtr->Pq2qg(zDn,hClu[ir],hDn,hR)/(y23*sijk)
            : 0.5*functionsPtr->Pg2gg(zDn,hClu[ir],hDn,hR)/(y23*sijk);
        }
        // Compare antenna and AP answers
        if (verbose >= 3) {
          cout<<"            "<<" hG ="<<setw(3)<<hClu[2]
              <<" GlbHel = "<<antGlbHel<<" ";
          cout<<" APhel = "<<PzAPupHel+PzAPdnHel<<endl;
        } 

	// Sum
	antGlbHelSum  += colFac * Pimp * Pari * antGlbHel * MEcluHel;
	PzAPhelSum    += colFac * (PordUp*PzAPupHel+PordDn*PzAPdnHel)*MEcluHel;

      }

    }

    if (verbose >= 2) {
      cout<<"\n    unpolarized   : glbSumSmooth  = "<<setw(10)<<antGlbSum;
      cout<<"   glbSumStrong = "<<setw(10)<<antGlbStrongSum;
      cout<<"   APsum  = "<<setw(10)<<PzAPsum<<endl;
      cout<<"    helicity test : glbSum  = "<<setw(10)<<antGlbHelSum;
      cout<<"   APsum  = "<<setw(10)<<PzAPhelSum;
      cout<<endl<<endl;
    } 
    
    //************************************************************************
    // Fill histograms
    //************************************************************************

    // Global
    double Rglb = antGlbSum/MGME;
    double lnRglb = (Rglb > 0.) ? log10(Rglb) : xLoR;
    double RglbStrong = antGlbStrongSum/MGME;
    //double lnRglbStrong = (Rglb > 0.) ? log10(RglbStrong) : xLoR;
    // DGLAP
    double Rap = PzAPsum/MGME;
    double lnRap = (Rap > 0.) ? log10(Rap) : xLoR;
    
    VRglb[0]->Fill(lnRglb,wRglb);
    VRap[0]->Fill(lnRap,wRap);
    double y12 = m2(pPartons[1]+pPartons[2])/S;
    //double y13 = m2(pPartons[1]+pPartons[3])/S;
    double y23 = m2(pPartons[2]+pPartons[3])/S;
    //double y24 = m2(pPartons[2]+pPartons[4])/S;
    //double y34 = m2(pPartons[3]+pPartons[4])/S;
    double y123 = m2(pPartons[1]+pPartons[2]+pPartons[3])/S;
    double y234 = m2(pPartons[2]+pPartons[3]+pPartons[4])/S;
    if ( y23 < 0.01 ) {
      VRglb[1]->Fill(lnRglb,wRglb);
      VRap[1]->Fill(lnRap,wRap);
      Vweights[1]->Fill(MGME,wWeights);
      double eg = pPartons[2].e() + pPartons[3].e();
      double xg = 2*eg/eCM;
      // Require branched gluon to be fairly hard and well-separated
      if (xg > 0.5 && min(y123,y234) > 0.25) {
        VRglb[2]->Fill(lnRglb,wRglb);
        VRap[2]->Fill(lnRap,wRap);
        Vweights[2]->Fill(MGME,wWeights);
      }
    }
    // 2D triple-collinear test    
    double lnx      = abs(log(y12/y123));
    double lny      = abs(log(y23/y123));
    Hn2D->Fill(lnx,lny);
    Hr2Dsmooth->Fill(lnx,lny,Rglb);
    Hr2Dstrong->Fill(lnx,lny,RglbStrong);

    // Periodically update displays
    vroot.update();       

  } // End loop over events
  
  // Last update/normalization
  Hr2Dsmooth = new TH2D((*Hr2Dsmooth)/(*Hn2D));  
  Hr2Dsmooth->SetMinimum(0.);
  Hr2Dsmooth->SetMaximum(2.);
  vroot.createDisplay("r2Dsmooth",Hr2Dsmooth,400,400);
  Hr2Dstrong = new TH2D((*Hr2Dstrong)/(*Hn2D));  
  Hr2Dstrong->SetMinimum(0.);
  Hr2Dstrong->SetMaximum(2.);
  vroot.createDisplay("r2Dstrong",Hr2Dstrong,400,400);
  vroot.update(2*nEvent);       

  //Write Vincia ROOT file
  cout<<"Saving root file... "<<rootFileName.str()<<endl;
  rootFile.Write();  
  rootFile.Close();

  // Press key to exit (and kill displays)
  cout << endl << "Enter q to terminate (will kill displays)..." << endl;
  char tmp;cin>>tmp;

  // Done.
  return 0;
}
