// Branchelemental.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2019 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions for the FSR & ISR Branchelementals class.

#include "Vincia/Branchelementals.h"

namespace Vincia {

//**********************************************************************

// BranchelementalFSR class member functions

//*********
// Simple print utility, showing the contents of the BranchelementalFSR
  
void BranchelementalFSR::list(bool header, bool footer) const { 
  if (header) {
    // Header.
    cout<< "\n --------  VINCIA FSR Dipole-Antenna Listing  --------------------------"
	<< "------------------------------------------------------------\n \n" 
	<< "  sys       antenna type     hels    col     iL   mothers   iR"
	<< "         m  Saved Trial Swap    Qtrial "
	<< " nVeto nHull nHadr\n"
	<< fixed << setprecision(3);
  }
  cout<< setw(5) << system<<" ";
  cout<< setw(9) << old1.name();
  cout<< setw(9) << old2.name()<<" ";
  cout<< setw(2) << "FF ";
  cout<< setw(2) << int(old1.pol()) <<" "<<setw(2)<< int(old2.pol())<<" ";
  cout<< setw(6) << old1.col()<<"  ";
  cout<< setw(5) << iL;
  cout<< setw(5) << iOld1 <<setw(5) << iOld2;
  cout<< setw(5) << iR ;
  cout<< num2str(m(),10);
  int iWin = getTrialIndex();
  if (nTrialGenerators() >= 1 && iWin >= 0) {
    if (hasSavedTrial[iWin]) {
      cout<< setw(13)<< trialGenPtrSav[iWin]->name();
      cout<< setw(5) << (isSwappedSav[iWin] ? "y" : "n");
      cout<< num2str(scaleSav[iWin],10);
    }
  } else {
    cout<<setw(13)<<"-";
    cout<<setw(15)<<" ";
  }
  cout<< " "<<setw(6)<<nVeto;   
  cout<< " "<<setw(5)<<nHull;   
  cout<< " "<<setw(5)<<nHadr;   
  cout<< endl;   
  if (footer) {
    cout<< "\n --------  End VINCIA VinciaShower Dipole-Antenna Listing  ------------"
       << "-------------------------------------------------------------\n";
  }
}

  //*********
  
  // Reset and populate list of (QCD) Trial Generators for this
  // Branchelemental.
  // Note: the names used must correspond to those returned
  // by the corresponding TrialGenerator.
  
  void BranchelementalFSR::resetTrialGenerators(
    map<string,TrialGeneratorFSR*> trialGenPtr, bool addConst) {

    // Reset
    clearTrialGenerators();   

    // Parent ID
    int id1      = old1.id();
    int id2      = old2.id();
    int colType1 = old1.colType();
    int colType2 = old2.colType();

    // Treat sextets as triplets here
    // Note: Generate 6-6bar with headRoom = 2
    // Note: Generate 3-3 with headRoom = 1/2 (but less important)
    if (colType1 ==  3) colType1 =  1;
    if (colType2 == -3) colType2 = -1;
    
    // Gluon emission trial antennae:
    // QQbar -> QGQbar
    if ( colType1 == 1 && colType2 == -1 ) {
      addTrialGenerator(iQQemit, false, trialGenPtr["TrialSoft"]);
      if (addConst)
        addTrialGenerator(iQQemit, false, trialGenPtr["TrialConstEmit"]);
    }
    // QG -> QGG
    else if ( colType1 == 1 && colType2 == 2 ) {
      addTrialGenerator(iQGemit, false, trialGenPtr["TrialSoft"]);
      if (addConst)
        addTrialGenerator(iQGemit, false, trialGenPtr["TrialConstEmit"]);
    }
    // GQbar -> GGQbar (physical antenna is obtained by swap of QG -> QGG)
    else if ( colType1 == 2 && colType2 == -1 ) { 
      addTrialGenerator(iQGemit, true, trialGenPtr["TrialSoft"]);
      if (addConst)
        addTrialGenerator(iQGemit, true, trialGenPtr["TrialConstEmit"]);
    }
    // GG -> GGG
    else if (colType1 == 2 && colType2 == 2) {
      addTrialGenerator(iGGemit, false, trialGenPtr["TrialSoft"]);
      if (addConst)
        addTrialGenerator(iGGemit, false, trialGenPtr["TrialConstEmit"]);
    }
    
    // Gluon splitting: gluon-collinear terms
    // GG -> GQbarQ + QbarQG
    if (id2 == 21 && id1 == 21) {
      addTrialGenerator(iGGsplit, false, trialGenPtr["TrialSplitK"]);
      addTrialGenerator(iGGsplit, true, trialGenPtr["TrialSplitI"]);
      if (addConst)
        addTrialGenerator(iGGemit, false, trialGenPtr["TrialConstSplit"]);
    }    
    // Everything else treated as if partner is quark
    else if (id2 == 21) {
      addTrialGenerator(iQGsplit, false, trialGenPtr["TrialSplitK"]);
      if (addConst)
        addTrialGenerator(iQGsplit, false, trialGenPtr["TrialConstSplit"]);
    } else if (id1 == 21) {
      addTrialGenerator(iQGsplit, true, trialGenPtr["TrialSplitI"]);
      if (addConst)
        addTrialGenerator(iQGsplit, true, trialGenPtr["TrialConstSplit"]);
    }
    
  }  

  
  //**********************************************************************
  
  // ISRBranchelemental class member functions

  //**********************************************************************
  // Simple print utility, showing the contents of the ISRBranchelemental

  void BranchelementalISR::list(bool header, bool footer) const { 
    if (header) {
      cout<< "\n --------  VINCIA ISR Dipole-Antenna Listing  ----------------------"
	  << "-----------  (S=sea, V=val, F=final)  --------------------------\n \n" 
	  << "  sys       antenna type     hels    col     iL   mothers   iR"
	  << "         m  TrialGenerators\n";
    }
    cout<< setw(5)<< system<<" ";
    cout<< setw(9) << old1.name();
    cout<< setw(9) << old2.name()<<" ";
    if (old1.isFinal())
      cout<<"F";
    else if (isVal1) cout<<"V";
    else cout<<"S";
    if (old2.isFinal())
      cout<<"F";
    else if (isVal2) cout<<"V";
    else cout<<"S";
    cout<<" ";
    cout<< setw(2) << int(old1.pol()) <<" "<<setw(2)<< int(old2.pol())<<" ";
    cout<< setw(6)<<col<<"  ";
    cout<< setw(5)<< iL;
    cout<< setw(5)<< iOld1;
    cout<< setw(5)<< iOld2;
    cout<< setw(5)<< iR;
    cout<<setw(10)<<m()<<" ";
    for (int j=0; j<(int)trialGenPtrSav.size(); j++) {
      string trialName = trialGenPtrSav[j]->name();
      trialName.erase(0,5);
      cout<< " "<<trialName;
    }
    cout<< "\n";  
    if (footer) {
      cout<< "\n --------  End VINCIA SpaceShower Antenna Listing  -----------------"
	  << "----------------------------------------------------------------\n";
    }
  }
  
}
